/** locate_vertices_in_other_map.c

For each vertex of map #im, determine which polygon of the other map it is in.

This uses global vars: im, verts, edges, n_verts, n_edges,
polygon_containing_cell, grid.

Method:

1. Run a ray up from the point and see whether it crosses any edges in the same
cell.  If so take the first one and return the appropriate adjacent polygon.
Note that the edge must cross the ray while it is in the current cell, or the
crossing doesn't count.

2. If the ray crosses no lines in this cell, then try the next cell up, and so
on.  If a cell is completely inside a polygon, then return that polygon.

3. If a cell being traversed has no edges from this map at all, not just none
crossing the ray, then when the polygon is determined, then label this cell as
being completely inside that polygon.

Wm. Randolph Franklin

Last modified: by Wm Randolph Franklin (wrf) on mab.ecse.rpi.edu at Thu Dec  5 19:02:32 1991

	%W%	%G%

*/

#include <stdio.h>
#include <math.h>
#include "overlay.h"

static int iv;			/* current vertex number */
static int cx, cy;		/* cell containing current vertex */
static int cy2, cy3;		/* current cell number */
static VERTEX_T *tv;		/* pointer to current vertex */
static EDGE_NO_T *tc;		/* pointer to current cell */
static int ne;			/* number edges (of other map) in current cell */
static int ipol;		/* current polygon */
static int im2;			/* the other map */
static int base;		/* base of edges in this cell */
static EDGE_T *te;		/* current edge */
static int ie;			/* current edge */
static VERT_COORD_T vx, vy;	/* current vertex */
static VERTEX_T *v1, *v2;	/* endpoints of current edge */
static VERT_COORD_T xl, xh;	/* of current edge */
static double xsecty;		/* Y coord of xsect of edge with vertical line from
				 * point */
static double besty;		/* Lowest xsecty seen so far */
double  diffy;			/* Vertical distance from point to edge. +ve =>
				 * point higher. */
static EDGE_T *beste;		/* best edge so far */
static VERT_COORD_T bestyh;
static int *nec;		/* number of edges in some cells of current column */
static  test_against_edges_of_this_cell();
static int plot_cell = 0;	/* Plot current cell ? */
static float fmin, fmax;	/* internal coord min and max */
static float glen;		/* size of a grid cell in internal units */
static float ttx, tty;		/* Total translation for debugging plot */
static float ts;		/* last scale " */


locate_vertices_in_other_map()
{

    im2 = 1 - im;
    fmin = INTERNAL_MIN;
    fmax = INTERNAL_MAX;
    glen = (fmax - fmin) / grid_size;

    plot_cell = 0;
    if (plot_cell) {
	if (im == 0)
	    start_graphics(fmin, fmax, fmin, fmax);
	draw_grid();

	/* Plot map 1 edges */

#ifdef USESUNVIEW
	set_linestyle(3);	/* dot dash */
#endif

	for (ie = 0; ie < n_edges[1]; ie++) {
	    te = &edges[1][ie];
	    v1 = &vertices[1][te->v1];
	    v2 = &vertices[1][te->v2];
#ifdef USESUNVIEW
	    move_abs_2((float) v1->x, (float) v1->y);
	    line_abs_2((float) v2->x, (float) v2->y);
#endif
	}

    }
    /*
     * Alloc nec only once even though locate_vertices_in_other_map is called twice.
     * This assumes locate_vertices_in_other_map is called first with im==0.
     */

    if (im == 0)
	nec = (int *)
	    malloc2((unsigned) grid_size * sizeof(int),
	    "Malloc failure trying to alloc nec in locate_vertices_in_other_map.");

    for (iv = 0; iv < n_verts[im]; iv++) {
	tv = &vertices[im][iv];	/* Pointer to this vertex's struct */
	vx = tv->x;
	vy = tv->y;
	cx = CELL(vx);
	cy = CELL(vy);

	/* for debugging */

	if (plot_cell && iv == 484) {
#ifdef USESUNVIEW
	    move_abs_2((float) vx, (float) vy);
	    set_linestyle(2);	/* dashed */
	    line_abs_2((float) vx, fmax);
#endif
	}
	for (cy2 = cy; cy2 <= grid_size; cy2++)	/* interate up the column of cells */
	    if (cy2 == grid_size)	/* fell off top of grid */
		ipol = 0;
	    else
	     /* Is whole cell in one polygon? */ if ((ipol = polygon_containing_cell[im2][cx][cy2]) != DONTKNOW)
		break;
	    else {
		tc = grid[cx][cy2];
		if (tc == NULL)
		    ne = 0;
		else {
		    if (im2 == 0)
			ne = tc[2] - (base = CELL_HEADER_LEN);	/* for map 0 */
		    else
			ne = tc[1] - (base = tc[2]);
		}

#ifdef USESUNVIEW
		/*
		 * Plot this cell.  The .75 is because the units are fractions of a
		 * NDC window and its y range is 0.75.  Messy.
		 */

		if (plot_cell && iv == 484) {
		    xseg((float) grid_size, -(float) cx, -0.75 * (float) cy2);
		    set_linestyle(0);	/* solid */

		    for (ie = 0; ie < ne; ie++) {
			te = &edges[im2][tc[base + ie]];
			v1 = &vertices[im2][te->v1];
			v2 = &vertices[im2][te->v2];
			move_abs_2((float) v1->x, (float) v1->y);
			line_abs_2((float) v2->x, (float) v2->y);
		    }
		}
#endif
		nec[cy2] = ne;
		if (ne == 0)
		    continue;	/* no edges in this cell */

		besty = HUGE_VAL;
		beste = NULL;
		test_against_edges_of_this_cell();

		if (beste == NULL)
		    continue;	/* no edges in this cell above this vertex */

		/*
		 * We have the edge!  Iff it is running left to right then we want
		 * the right polygon.
		 */

		if (vertices[im2][beste->v1].x < vertices[im2][beste->v2].x)
		    ipol = beste->pr;
		else
		    ipol = beste->pl;

		break;
	    }

	tv->p = ipol;
	for (cy3 = cy2 - 1; cy3 >= cy; cy3--) {	/* Iterate down labelling cells when
						 * possible. */
	    if (polygon_containing_cell[im2][cx][cy3] != DONTKNOW)
		break;		/* Stop at first cell where we know the containing
				 * polygon. */
	    if (nec[cy3] > 0)
		break;		/* Stop if cell has edges; then it can't be
				 * completely in one polygon. */
	    polygon_containing_cell[im2][cx][cy3] = ipol;
	}
    }				/* for(iv=0... */
}


static
test_against_edges_of_this_cell()
{

    int     flag;

    for (ie = 0; ie < ne; ie++) {
	te = &edges[im2][tc[base + ie]];
	v1 = &vertices[im2][te->v1];
	v2 = &vertices[im2][te->v2];
	xl = MIN(v1->x, v2->x);
	xh = MAX(v1->x, v2->x);

	/*
	 * Is point outside x bounds of this edge?  Use simulation of simplicity:
	 * shift map 1 by epsilon relative to map 0.  This also eliminates vertical
	 * edges.
	 */

	if (vx < xl || vx > xh || vx == ((im == 0) ? xl : xh))
	    continue;

	/* Is point above edge?  If point is on edge, use simulation of simplicity. */

	xsecty = ((double) (-te->a * vx - te->c)) / (double) (te->b);
	diffy = vy - xsecty;

	if (diffy == 0)
	    diffy = ((im == 0) ? -te->a : te->a);
	if (diffy == 0)
	    diffy = ((im == 0) ? -te->b : te->b);
	if (diffy == 0)
	    myerror("Zero length edge in locate_vertices_in_other_map");

	if (diffy > 0)
	    continue;		/* Yes, point is above edge. */

	if (xsecty > besty)
	    continue;		/* Point is below, but another edge is closer. */

	if (xsecty == besty) {	/* Line up from point cuts this edge and another
				 * edge at the same Y.  This should only happen at a
				 * vertex.  Roundoff error should not cause it, I
				 * hope.  If im==0, we're at the right end of both
				 * edges.  im==1 => left end.  */

	    /* flag:  Is this slope algebraically less than the last one? */

	    flag = (((double) (te->a) / (double) (te->b)) > ((double) (beste->a) / (double) (beste->b)));

	    /* If im==0 we want the bigger slope, if im==1, the smaller. */

	    if ((im && !flag) || (flag && !im))
		continue;

	}
	besty = xsecty;		/* We have a new closer edge.  Update out records. */
	beste = te;
	bestyh = MAX(v1->y, v2->y);
    }

    /*
     * Check that the best edge intersects the ray within the current cell.  If not
     * ignore it and act as if there were no edges in this cell crossing the ray. If
     * the hi end of the edge is in the cell then the intersection point certainly
     * is.
     */

    if (CELL(bestyh) == cy2)
	return;
    if (CELL(besty) > cy2) {
	/* printf("Intersection of edge and ray is not in cell %d, %d\n", cx, cy2); */
	beste = NULL;
    }
}


/* Transform segment 1.  SunCore manual p. 89 */

xseg(s, tx, ty)
    float   s, tx, ty;
{
#ifdef USESUNVIEW
    ts = s;
    ttx = tx;
    tty = ty;
    set_segment_image_transformation_2(1, s, s, 0.0, tx, ty);
#endif
}

/* Incrementally shift the plot */

xi(d)
    float   d;
{
#ifdef USESUNVIEW
    ttx += d;
    set_segment_image_transformation_2(1, ts, ts, 0.0, ttx, tty);
    printf("ttx=%f, tty=%f\n", ttx, tty);
#endif
}

yi(d)
    float   d;
{
#ifdef USESUNVIEW
    tty += d;
    set_segment_image_transformation_2(1, ts, ts, 0.0, ttx, tty);
    printf("ttx=%f, tty=%f\n", ttx, tty);
#endif
}




/* Draw the grid dotted */

draw_grid()
{
    int     i;
    float   d;
    char    s[100];

#ifdef USESUNVIEW
    set_linestyle(1);		/* dotted */

    for (i = 0; i <= grid_size; i++) {
	d = fmin + i * glen;
	move_abs_2(d, fmin);
	line_abs_2(d, fmax);
	move_abs_2(fmin, d);
	line_abs_2(fmax, d);
	sprintf(s, "%d", i);
	move_abs_2(d, fmin + glen / 2.);
	text(s);
	move_abs_2(fmin + glen / 2., d);
	text(s);
    }
#endif
}

int
cell(xy)
    VERT_COORD_T xy;
{
    return (CELL(xy));
}
